package com.novayre.jidoka.robot.tutorial;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.exception.ExceptionUtils;

import com.novayre.jidoka.client.api.IJidokaServer;
import com.novayre.jidoka.client.api.IRobot;
import com.novayre.jidoka.client.api.JidokaFactory;
import com.novayre.jidoka.client.api.annotations.Robot;
import com.novayre.jidoka.client.api.exceptions.JidokaFatalException;
import com.novayre.jidoka.client.api.exceptions.JidokaItemException;
import com.novayre.jidoka.client.api.exceptions.JidokaUnsatisfiedConditionException;
import com.novayre.jidoka.client.api.multios.EClientShowWindowType;
import com.novayre.jidoka.client.api.multios.IClient;

/**
 * Notepad robot.
 * 
 * @author Jidoka
 *
 */
@Robot
public class NotepadRobot implements IRobot {

	/**
	 * Regular expression to search the window for the application Notepad. It
	 * changes depending on the platform language.
	 */
	private static final String NOTEPAD_REGEXP = ".*Notepad";

	/**
	 * Default pause between actions imitating human behavior.
	 */
	private static final int PAUSE = 500;

	/**
	 * Items to process.
	 */
	private static final String[] ITEMS = new String[] { "This is the \"Hello World!\" Jidoka Robot",
			"It shows how to write into the notepad" };

	/**
	 * Server.
	 */
	private IJidokaServer<?> server;

	/**
	 * Client module.
	 */
	private IClient client;

	/**
	 * Current item index. The first one is the number '1'.
	 */
	private int currentItemIndex = 1;

	@Override
	public boolean startUp() throws Exception {
		// Initialization of the robot components
		server = (IJidokaServer<?>) JidokaFactory.getServer();
		client = IClient.getInstance(this);

		return IRobot.super.startUp();
	}

	/**
	 * Initial action 'Init'.
	 */
	public void init() {

		try {
			// Default pause after typing or using the mouse
			client.typingPause(PAUSE);
			client.mousePause(PAUSE);

			// Set the number of items
			server.setNumberOfItems(ITEMS.length);
		} catch (Exception e) {
			throw new JidokaFatalException("Error initializing the robot");
		}
	}

	/**
	 * Action 'Open Notepad'.
	 * <p>
	 * First way to open Notepad.
	 */
	public void openNotepadAlt1() {

		try {
			// Win+R
			client.keyboard().windows("r").pause().type("notepad").enter().pause();

			// Wait during 10 seconds until the application window is present
			if (client.waitFor(this).window(10, NOTEPAD_REGEXP) == null) {

				/*
				 * If the window is not present at the end of the wait, the robot must fail.
				 */
				throw new JidokaFatalException("Notepad could not be opened");
			}

			// Maximize the window
			client.showWindow(client.getWindow(NOTEPAD_REGEXP).getId(), EClientShowWindowType.MAXIMIZE);
		} catch (Exception e) {
			throw new JidokaFatalException("Error opening notepad");
		}
	}

	/**
	 * Action 'Open Notepad'.
	 * <p>
	 * Second way to open Notepad.
	 * 
	 * @throws JidokaUnsatisfiedConditionException if some conditions are not met
	 */
	public void openNotepadAlt2() throws JidokaUnsatisfiedConditionException {

		try {
			// Win+R
			client.keyboard().windows("r").pause().type("notepad").enter().pause();

			// Using a custom condition for waiting
			client.waitCondition(10, 1000, "Wait for Notepad to be opened", null, true,
					(i, c) -> client.getWindow(NOTEPAD_REGEXP) != null);

			// Maximize the window
			client.showWindow(client.getWindow(NOTEPAD_REGEXP).getId(), EClientShowWindowType.MAXIMIZE);
		} catch (Exception e) {
			throw new JidokaFatalException("Error opening notepad");
		}
	}

	/**
	 * Action 'Process line'.
	 */
	public void processItem() {

		try {
			// Get the current item
			String item = ITEMS[currentItemIndex - 1];

			/*
			 * Notify the server the item treatment starts. Important to calculate
			 * statistics and execution times. The second parameter is used to make it
			 * easier to find registers in the execution log.
			 */
			server.setCurrentItem(currentItemIndex, item);

			// Type the string into de Notepad and press Enter
			client.keyboard().type(item).pause().enter();

			/*
			 * Notify the item result. In this case, a the detail message is not specified.
			 */
			server.setCurrentItemResultToOK();
		} catch (Exception e) {
			throw new JidokaItemException("Error processing the item");
		}
	}

	/**
	 * Action 'More?'.
	 * 
	 * @return the wire name
	 */
	public String hasMoreItems() {

		try {
			if (currentItemIndex == ITEMS.length) {
				// All items have been processed
				return "no";
			}

			// There are more items to process
			currentItemIndex++;

			return "yes";
		} catch (Exception e) {
			throw new JidokaItemException("Error evaluating item");
		}
	}

	/**
	 * Action 'End'.
	 */
	public void end() {
		// Continue the process. At this step, the robot ends its execution
	}

	/**
	 * @see com.novayre.jidoka.client.api.IRobot#cleanUp()
	 */
	@Override
	public String[] cleanUp() throws Exception {

		// Close all Notepad instances
		client.killAllProcesses("notepad.exe", 500);

		return new String[0];
	}
	
	/**
	 * Manage exception.
	 *
	 * @param action    the action
	 * @param exception the exception
	 * @return the string
	 * @throws Exception the exception
	 */
	@Override
	public String manageException(String action, Exception exception) throws Exception {

		// We get the message of the exception
		String errorMessage = ExceptionUtils.getRootCause(exception).getMessage();

		// We send a screenshot to the log so the user can see the screen in the moment
		// of the error
		// This is a very useful thing to do
		server.sendScreen("Screenshot at the moment of the error");

		// If we have a FatalException we should abort the execution.
		if (ExceptionUtils.indexOfThrowable(exception, JidokaFatalException.class) >= 0) {

			server.error(StringUtils.isBlank(errorMessage) ? "Fatal error" : errorMessage);
			return IRobot.super.manageException(action, exception);
		}

		// If the error is processing one items we must mark it as a warning and go on
		// with the next item
		if (ExceptionUtils.indexOfThrowable(exception, JidokaItemException.class) >= 0) {

			server.warn(StringUtils.isBlank(errorMessage) ? "Item error" : errorMessage);
			server.setCurrentItemResultToWarn(errorMessage);
			return "hasMoreItems";
		}

		server.warn("Unknown exception!");

		// If we have any other exception we must abort the execution, we don't know
		// what has happened

		return IRobot.super.manageException(action, exception);
	}
}
